"""Unit tests of mesh_smooth."""
import pickle

import numpy as np

from brainstat.mesh.data import mesh_smooth

from .testutil import datadir


def dummy_test(infile, expfile):

    # load input test data
    ifile = open(infile, "br")
    idic = pickle.load(ifile)
    ifile.close()

    Y = idic["Y"]
    FWHM = idic["FWHM"]

    surf = {}
    if "tri" in idic.keys():
        surf["tri"] = idic["tri"]

    if "lat" in idic.keys():
        surf["lat"] = idic["lat"]

    # run mesh_smooth
    Y_out = mesh_smooth(Y, surf, FWHM)

    # load expected outout data
    efile = open(expfile, "br")
    expdic = pickle.load(efile)
    efile.close()
    Y_exp = expdic["Python_Y"]

    testout = []

    comp = np.allclose(Y_out, Y_exp, rtol=1e-05, equal_nan=True)
    testout.append(comp)

    assert all(flag == True for (flag) in testout)


def test_01():
    infile = datadir("xstatsmo_01_IN.pkl")
    expfile = datadir("xstatsmo_01_OUT.pkl")
    dummy_test(infile, expfile)



"""Unit tests of fdr."""

import pickle

import numpy as np
import pytest

from brainstat.stats._multiple_comparisons import _fdr
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect
from brainstat.tests.testutil import datadir


def dummy_test(infile, expfile):

    # load input test data
    ifile = open(infile, "br")
    idic = pickle.load(ifile)
    ifile.close()

    slm = SLM(FixedEffect(1), FixedEffect(1))
    for key in idic.keys():
        setattr(slm, key, idic[key])

    # run fdr
    Q = _fdr(slm)

    # load expected outout data
    efile = open(expfile, "br")
    expdic = pickle.load(efile)
    efile.close()

    assert np.allclose(Q, expdic["Q"])


expected_number_of_tests = 14
parametrize = pytest.mark.parametrize


@parametrize("test_number", range(1, expected_number_of_tests + 1))
def test_run_all(test_number):
    infile = datadir("xstatq_" + f"{test_number:02d}" + "_IN.pkl")
    expfile = datadir("xstatq_" + f"{test_number:02d}" + "_OUT.pkl")
    dummy_test(infile, expfile)



"""Data generation for mesh_edges unit tests."""

import pickle

import numpy as np

from brainstat.mesh.utils import mesh_edges
from brainstat.tests.testutil import datadir


def generate_random_mesh_edge_data(
    key_dim,
    finname,
    key_name="tri",
    key_dtype="float",
    seed=0,
):
    """Generate random test datasets."""
    # key_dim : tuple, dimensions of array
    # finname : string, filename ending with *pkl
    # key_name : string, "tri" or "lat"
    # key_dtype : string, float or int

    np.random.seed(seed=seed)

    surf = {}

    if key_name == "tri":
        surf["tri"] = np.random.randint(1, 1024, size=key_dim)
    elif key_name == "lat":
        surf["lat"] = np.random.randint(0, 2, size=key_dim)
    surf[key_name] = surf[key_name].astype(key_dtype)

    with open(finname, "wb") as handle:
        pickle.dump(surf, handle, protocol=4)
    return surf


def get_meshedge_output(surf, foutname):
    """Runs mesh_edges and returns all relevant output."""

    # run mesh_edges
    surf_out = {}
    surf_out["edg"] = mesh_edges(surf)

    with open(foutname, "wb") as handle:
        pickle.dump(surf_out, handle, protocol=4)  #

    return


def generate_data_test_mesh_edges():

    ### test_01 data in-out generation
    print("test_mesh_edges.py : test_01 data is generated..")
    # ['tri'] is a 2D numpy array of shape (78, 3), dtype('float64')
    key_dim = (78, 3)
    finname = datadir("xstatedg_01_IN.pkl")
    key_name = "tri"
    key_dtype = np.float64
    seed = 444
    surf = generate_random_mesh_edge_data(key_dim, finname, key_name, key_dtype, seed)
    foutname = datadir("xstatedg_01_OUT.pkl")
    get_meshedge_output(surf, foutname)

    ### test_02 data in-out generation
    print("test_mesh_edges.py : test_02 data is generated..")
    # ['lat'] is a 2D numpy array of shape (10, 10), dtype('float64')
    key_dim = (10, 10)
    finname = datadir("xstatedg_02_IN.pkl")
    key_name = "lat"
    key_dtype = np.float64
    seed = 445
    surf = generate_random_mesh_edge_data(key_dim, finname, key_name, key_dtype, seed)
    foutname = datadir("xstatedg_02_OUT.pkl")
    get_meshedge_output(surf, foutname)

    ### test_03 data in-out generation
    print("test_mesh_edges.py : test_03 data is generated..")
    # ['lat'] is a 3D numpy array of shape (10, 10, 10), dtype('int64')
    key_dim = (10, 10, 10)
    finname = datadir("xstatedg_03_IN.pkl")
    key_name = "lat"
    key_dtype = np.int64
    seed = 446
    surf = generate_random_mesh_edge_data(key_dim, finname, key_name, key_dtype, seed)
    foutname = datadir("xstatedg_03_OUT.pkl")
    get_meshedge_output(surf, foutname)

    ### test_04 data in-out generation
    print("test_mesh_edges.py : test_04 data is generated..")
    # ['tri'] is a 2D numpy array of shape (2044, 3), dtype('uint16')
    key_dim = (2044, 3)
    finname = datadir("xstatedg_04_IN.pkl")
    key_name = "tri"
    key_dtype = np.uint16
    seed = 447
    surf = generate_random_mesh_edge_data(key_dim, finname, key_name, key_dtype, seed)
    foutname = datadir("xstatedg_04_OUT.pkl")
    get_meshedge_output(surf, foutname)


if __name__ == "__main__":
    generate_data_test_mesh_edges()



"""Operations on data on a mesh."""
from typing import Union

import numpy as np
from brainspace.vtk_interface.wrappers.data_object import BSPolyData

from .utils import mesh_edges


def mesh_smooth(
    Y: np.ndarray, surf: Union[dict, BSPolyData], FWHM: float
) -> np.ndarray:
    """Smooths surface data by repeatedly averaging over edges.

    Parameters
    ----------
    Y : numpy.ndarray
        Surface data of shape (n,v) or (n,v,k). v is the number of vertices,
        n is the number of observations, k is the number of variates.
    surf : dict, BSPolyData
        A dictionary with key 'tri' or 'lat', or a BSPolyData object of the
        surface. surf['tri'] is a numpy.ndarray  of shape (t,3), t is the
        triangle indices, or surf['lat'] is a numpy.ndarray of shape (nx,ny,nz),
        where (nx,ny,nz) is the volume size, values are 1=in, 0=out.
    FWHM : float
       Gaussian smoothing filter in mesh units.

    Returns
    -------
    numpy.ndarray
        Smoothed surface data of shape (n,v) or (n,v,k).
    """

    niter = int(np.ceil(pow(FWHM, 2) / (2 * np.log(2))))
    if isinstance(Y, np.ndarray):
        Y = np.array(Y, dtype="float")
        if np.ndim(Y) == 2:
            n, v = np.shape(Y)
            k = 1
            isnum = True

        elif np.ndim(Y) == 3:
            n, v, k = np.shape(Y)
            isnum = True

    edg = mesh_edges(surf) + 1
    agg_1 = np.bincount(edg[:, 0], minlength=(v + 1)) * 2
    agg_2 = np.bincount(edg[:, 1], minlength=(v + 1)) * 2
    Y1 = (agg_1 + agg_2)[1:]

    if n > 1:
        print(" %i x %i surfaces to smooth, %% remaining: 100 " % (n, k))

    n10 = np.floor(n / 10)

    for i in range(0, n):

        if n10 != 0 and np.remainder(i + 1, n10) == 0:
            print("%s " % str(int(100 - (i + 1) / n10 * 10)), end="")

        for j in range(0, k):
            if isnum:
                if np.ndim(Y) == 2:
                    Ys = Y[i, :]

                elif np.ndim(Y) == 3:
                    Ys = Y[i, :, j]

                for itera in range(1, niter + 1):
                    Yedg = Ys[edg[:, 0] - 1] + Ys[edg[:, 1] - 1]
                    agg_tmp1 = np.bincount(edg[:, 0], Yedg, (v + 1))[1:]
                    agg_tmp2 = np.bincount(edg[:, 1], Yedg, (v + 1))[1:]
                    with np.errstate(invalid="ignore"):
                        Ys = (agg_tmp1 + agg_tmp2) / Y1

                if np.ndim(Y) == 2:
                    Y[i, :] = Ys

                elif np.ndim(Y) == 3:
                    Y[i, :, j] = Ys
    if n > 1:
        print("Done")

    return Y



"""Operations on meshes."""

import sys
from typing import TYPE_CHECKING, Optional, Tuple, Union

import numpy as np
from brainspace.mesh.mesh_elements import get_edges
from brainspace.vtk_interface.wrappers.data_object import BSPolyData

from brainstat._typing import ArrayLike
from brainstat.stats.utils import colon

if TYPE_CHECKING:
    from brainstat.stats.SLM import SLM  # type: ignore


def mesh_edges(
    surf: Union[dict, BSPolyData, "SLM"], mask: Optional[ArrayLike] = None
) -> np.ndarray:
    """Converts the triangles or lattices of a mesh to edges.

    Args:
        surf (dict): = a dictionary with key 'tri' or 'lat'
        surf['tri'] = (t x 3) numpy array of triangle indices, t:#triangles, or,
        surf['lat'] = 3D numpy array of 1's and 0's (1:in, 0:out).
        or
        surf (BSPolyData) = a BrainSpace surface object
        or
        surf (SLM) = a SLM object with an associated surface.

    Returns:
        edg (np.array): A e-by-2 numpy array containing the indices of the edges, where
        e is the number of edges.
    """

    # This doesn't strictly test that its BrainStat SLM, but we can't import
    # directly without causing a circular import.
    if not isinstance(surf, dict) and not isinstance(surf, BSPolyData):
        if surf.tri is not None:
            surf = {"tri": surf.tri}
        elif surf.lat is not None:
            surf = {"lat": surf.lat}
        elif surf.surf is not None:
            return mesh_edges(surf.surf)
        else:
            ValueError("SLM object does not have triangle/lattice data.")

    # For BSPolyData, simply use BrainSpace's functionality to grab edges.
    if isinstance(surf, BSPolyData):
        edg = get_edges(surf)

    # Convert triangles to edges by grabbing all unique edges within triangles.
    elif "tri" in surf:
        tri = np.sort(surf["tri"], axis=1)
        edg = np.unique(
            np.concatenate(
                (np.concatenate((tri[:, [0, 1]], tri[:, [0, 2]])), tri[:, [1, 2]])
            ),
            axis=0,
        )
        edg = edg - 1

    elif "lat" in surf:
        # See the comments of SurfStatResels for a full explanation.
        if surf["lat"].ndim == 2:
            surf["lat"] = np.expand_dims(surf["lat"], axis=2)

        I, J, K = np.shape(surf["lat"])
        IJ = I * J

        a = np.arange(1, int(I) + 1, dtype="int")
        b = np.arange(1, int(J) + 1, dtype="int")

        i, j = np.meshgrid(a, b)
        i = i.T.flatten("F")
        j = j.T.flatten("F")

        n1 = (I - 1) * (J - 1) * 6 + (I - 1) * 3 + (J - 1) * 3 + 1
        n2 = (I - 1) * (J - 1) * 3 + (I - 1) + (J - 1)

        edg = np.zeros(((K - 1) * n1 + n2, int(2)), dtype="int")

        for f in range(0, 2):

            c1 = np.where((np.remainder((i + j), 2) == f) & (i < I) & (j < J))[0]
            c2 = np.where((np.remainder((i + j), 2) == f) & (i > 1) & (j < J))[0]
            c11 = np.where((np.remainder((i + j), 2) == f) & (i == I) & (j < J))[0]
            c21 = np.where((np.remainder((i + j), 2) == f) & (i == I) & (j > 1))[0]
            c12 = np.where((np.remainder((i + j), 2) == f) & (i < I) & (j == J))[0]
            c22 = np.where((np.remainder((i + j), 2) == f) & (i > 1) & (j == J))[0]

            # bottom slice
            edg0 = (
                np.block(
                    [
                        [c1, c1, c1, c2 - 1, c2 - 1, c2, c11, c21 - I, c12, c22 - 1],
                        [
                            c1 + 1,
                            c1 + I,
                            c1 + 1 + I,
                            c2,
                            c2 - 1 + I,
                            c2 - 1 + I,
                            c11 + I,
                            c21,
                            c12 + 1,
                            c22,
                        ],
                    ]
                ).T
                + 1
            )
            # between slices
            edg1 = (
                np.block(
                    [
                        [c1, c1, c1, c11, c11, c12, c12],
                        [
                            c1 + IJ,
                            c1 + 1 + IJ,
                            c1 + I + IJ,
                            c11 + IJ,
                            c11 + I + IJ,
                            c12 + IJ,
                            c12 + 1 + IJ,
                        ],
                    ]
                ).T
                + 1
            )

            edg2 = (
                np.block(
                    [
                        [c2 - 1, c2, c2 - 1 + I, c21 - I, c21, c22 - 1, c22],
                        [
                            c2 - 1 + IJ,
                            c2 - 1 + IJ,
                            c2 - 1 + IJ,
                            c21 - I + IJ,
                            c21 - I + IJ,
                            c22 - 1 + IJ,
                            c22 - 1 + IJ,
                        ],
                    ]
                ).T
                + 1
            )

            if f:
                for k in colon(2, K - 1, 2):
                    edg[(k - 1) * n1 + np.arange(0, n1), :] = (
                        np.block([[edg0], [edg2], [edg1], [IJ, 2 * IJ]]) + (k - 1) * IJ  # type: ignore
                    )

            else:
                for k in colon(1, K - 1, 2):
                    edg[(k - 1) * n1 + np.arange(0, n1), :] = (
                        np.block([[edg0], [edg1], [edg2], [IJ, 2 * IJ]]) + (k - 1) * IJ  # type: ignore
                    )

            if np.remainder((K + 1), 2) == f:
                # top slice
                edg[(K - 1) * n1 + np.arange(0, n2), :] = (
                    edg0[np.arange(0, n2), :] + (K - 1) * IJ
                )

        # index by voxels in the "lat"
        vid = np.array(
            np.multiply(
                np.cumsum(surf["lat"][:].T.flatten()), surf["lat"][:].T.flatten()
            ),
            dtype="int",
        )
        vid = vid.reshape(len(vid), 1)

        # only inside the lat
        all_idx = np.all(
            np.block(
                [
                    [surf["lat"].T.flatten()[edg[:, 0] - 1]],
                    [surf["lat"].T.flatten()[edg[:, 1] - 1]],
                ]
            ).T,
            axis=1,
        )

        edg = vid[edg[all_idx, :] - 1].reshape(np.shape(edg[all_idx, :] - 1))
        edg = edg - 1

    else:
        sys.exit('Input "surf" must have "lat" or "tri" key, or be a mesh object.')

    if mask is not None:
        edg, _ = _mask_edges(edg, mask)

    return edg


def _mask_edges(edges: np.ndarray, mask: ArrayLike) -> Tuple[np.ndarray, np.ndarray]:
    # TODO: this section is sloppily written.
    missing_edges = np.where(np.logical_not(mask))
    remove_edges = np.zeros(edges.shape, dtype=bool)
    for i in range(edges.shape[0]):
        for j in range(edges.shape[1]):
            remove_edges[i, j] = (edges[i, j] == missing_edges).any()
    idx = ~np.any(remove_edges, axis=1)
    edges = edges[idx, :]
    edges = _make_contiguous(edges)
    return edges, idx


def _make_contiguous(Y: np.ndarray) -> np.ndarray:
    """Makes values of Y contiguous integers

    Parameters
    ----------
    Y : numpy.ndarray
        Array with uncontiguous numbers.

    Returns
    -------
    numpy.ndarray
        Array Y converted to contiguous numbers in range(np.unique(Y).size).
    """
    val = np.unique(Y)
    for i in range(val.size):
        Y[Y == val[i]] = i
    return Y



#!/usr/bin/env python3

"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
"""

import setuptools

from brainstat import __version__ as brainstat_version

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="brainstat",
    version=brainstat_version,
    author="MNI-MICA Lab and MPI-CNG Lab",
    author_email="reinder.vosdewael@gmail.com, sheymaba@gmail.com",
    description="A toolbox for statistical analysis of neuroimaging data",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/MICA-LAB/BrainStat",
    packages=setuptools.find_packages(),
    license="BSD 3-Clause License",
    package_data={
        "": ["logging.conf", "data_urls.json"],
    },
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.7.*",
    install_requires=[
        "abagen>=0.1",
        "brainspace>=0.1.2",
        "h5py",
        "netneurotools",
        "neurosynth",
        "nibabel",
        "nilearn>=0.7.0",
        "nimare",
        "numpy>=1.16.5",
        "pandas",
        "scikit_learn",
        "scipy>=1.3.3",
        "trimesh",
    ],
    extras_require={"dev": ["gitpython", "hcp-utils", "mypy", "plotly", "pytest"]},
    project_urls={  # Optional
        "Documentation": "https://brainstat.readthedocs.io",
        "Bug Reports": "https://github.com/MICA-LAB/BrainStat/issues",
        "Source": "https://github.com/MICA-LAB/BrainStat/",
    },
)



""" Meta-analytic decoding based on NiMARE """
import re
import urllib
import zipfile
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Generator, Optional, Sequence, Union

import nibabel as nib
import numpy as np
import pandas as pd
from brainspace.vtk_interface.wrappers.data_object import BSPolyData
from nilearn.datasets import load_mni152_brain_mask
from scipy.stats.stats import pearsonr

from brainstat._utils import data_directories, logger, read_data_fetcher_json
from brainstat.mesh.interpolate import multi_surface_to_volume


def surface_decoder(
    pial: Union[str, BSPolyData, Sequence[Union[str, BSPolyData]]],
    white: Union[str, BSPolyData, Sequence[Union[str, BSPolyData]]],
    stat_labels: Union[str, np.ndarray, Sequence[Union[str, np.ndarray]]],
    *,
    interpolation: str = "linear",
    data_dir: Optional[Union[str, Path]] = None,
    database: str = "neurosynth",
) -> pd.DataFrame:
    """Meta-analytic decoding of surface maps using NeuroSynth or NeuroQuery.

    Parameters
    ----------
    pial : str, BSPolyData, sequence of str or BSPolyData
        Path of a pial surface file, BSPolyData of a pial surface or a list
        containing multiple of the aforementioned.
    white : str, BSPolyData, sequence of str or BSPolyData
        Path of a white matter surface file, BSPolyData of a pial surface or a
        list containing multiple of the aforementioned.
    stat_labels : str, numpy.ndarray, sequence of str or numpy.ndarray
        Path to a label file for the surfaces, numpy array containing the
        labels, or a list containing multiple of the aforementioned.
    mask_labels : str, numpy.ndarray, sequence of str of or numpy.ndarray
        Path to a mask file for the surfaces, numpy array containing the
        mask, or a list containing multiple of the aforementioned. If None
        all vertices are included in the mask. Defaults to None.
    interpolation : str, optional
        Either 'nearest' for nearest neighbor interpolation, or 'linear'
        for trilinear interpolation, by default 'linear'.
    data_dir : str, optional
        The directory of the dataset. If none exists, a new dataset will
        be downloaded and saved to this path. If None, the directory defaults to
        your home directory, by default None.


    Returns
    -------
    pandas.DataFrame
        Table with correlation values for each feature.
    """

    data_dir = Path(data_dir) if data_dir else data_directories["NEUROSYNTH_DATA_DIR"]
    data_dir.mkdir(exist_ok=True, parents=True)

    logger.info(
        "Fetching Neurosynth feature files. This may take several minutes if you haven't downloaded them yet."
    )
    feature_files = tuple(_fetch_precomputed(data_dir, database=database))

    mni152 = load_mni152_brain_mask()

    with NamedTemporaryFile(suffix=".nii.gz", delete=False) as f:
        name = f.name
    try:
        multi_surface_to_volume(
            pial=pial,
            white=white,
            volume_template=mni152,
            output_file=name,
            labels=stat_labels,
            interpolation=interpolation,
        )

        stat_volume = nib.load(name)

        mask = (stat_volume.get_fdata() != 0) & (mni152.get_fdata() != 0)
        stat_vector = stat_volume.get_fdata()[mask]
    finally:
        Path(name).unlink()

    feature_names = []
    correlations = np.zeros(len(feature_files))

    logger.info("Running correlations with all Neurosynth features.")
    for i in range(len(feature_files)):
        feature_names.append(re.search("__[A-Za-z0-9]+", feature_files[i].stem)[0][2:])  # type: ignore
        feature_data = nib.load(feature_files[i]).get_fdata()[mask]
        keep = np.logical_not(
            np.isnan(feature_data)
            | np.isinf(feature_data)
            | np.isnan(stat_vector)
            | np.isinf(stat_vector)
        )
        correlations[i], _ = pearsonr(stat_vector[keep], feature_data[keep])

    df = pd.DataFrame(correlations, index=feature_names, columns=["Pearson's r"])
    return df.sort_values(by="Pearson's r", ascending=False)


def _fetch_precomputed(data_dir: Path, database: str) -> Generator[Path, None, None]:
    """Wrapper for any future data fetcher.

    Parameters
    ----------
    data_dir : Path
        Directory where the data is stored.
    database : str
        Name of the database, valid arguments are 'neurosynth'.

    Returns
    -------
    generator
        Generator of paths to the precomputed files.

    Raises
    ------
    NotImplementedError
        Returned when requesting the Neuroquery data fetcher.
    ValueError
        Returned when requesting an unknown database.
    """
    if database == "neurosynth":
        return _fetch_precomputed_neurosynth(data_dir)
    elif database == "neuroquery":
        raise NotImplementedError("Neuroquery has not been implemented yet.")
    else:
        raise ValueError(f"Unknown database {database}.")


def _fetch_precomputed_neurosynth(data_dir: Path) -> Generator[Path, None, None]:
    """Downloads precomputed Neurosynth features and returns the filepaths."""

    json = read_data_fetcher_json()["neurosynth_precomputed"]
    url = json["url"]

    existing_files = data_dir.glob("Neurosynth_TFIDF__*z_desc-consistency.nii.gz")

    if len(list(existing_files)) != json["n_files"]:
        logger.info("Downloading Neurosynth data files.")
        response = urllib.request.urlopen(url)

        # Open, close, and reopen file to deal with Windows permission issues.
        with NamedTemporaryFile(prefix=str(data_dir), suffix=".zip", delete=False) as f:
            name = f.name
        try:
            with open(name, "wb") as fw:
                fw.write(response.read())

            with zipfile.ZipFile(name, "r") as fr:
                fr.extractall(data_dir)
        finally:
            (Path(name)).unlink()

    return data_dir.glob("Neurosynth_TFIDF__*z_desc-consistency.nii.gz")



"""Unit tests of t-test."""
import pickle

import numpy as np
import pytest

from brainstat.stats.SLM import SLM
from brainstat.tests.testutil import array2effect, datadir


def dummy_test(infile, expfile):

    # load input test data
    ifile = open(infile, "br")
    idic = pickle.load(ifile)
    ifile.close()

    model = array2effect(idic["M"], idic["n_random"])
    contrast = -idic["M"][:, -1]

    # run _t_test
    slm = SLM(model, contrast, idic["surf"])
    slm._linear_model(idic["Y"])
    slm._t_test()

    # load expected outout data
    efile = open(expfile, "br")
    expdic = pickle.load(efile)
    efile.close()

    testout = []
    for key in expdic.keys():
        if isinstance(expdic[key], dict):
            slm_sub_dict = getattr(slm, key)
            exp_sub_dict = expdic[key]
            comp = np.all(
                [
                    np.allclose(slm_sub_dict[x], exp_sub_dict[x])
                    for x in exp_sub_dict.keys()
                ]
            )
        else:
            comp = np.allclose(
                getattr(slm, key), expdic[key], rtol=1e-05, equal_nan=True
            )
            testout.append(comp)

    assert all(flag == True for (flag) in testout)


expected_number_of_tests = 16
parametrize = pytest.mark.parametrize


@parametrize("test_number", range(1, expected_number_of_tests + 1))
def test_run_all(test_number):
    infile = datadir("xstatt_" + f"{test_number:02d}" + "_IN.pkl")
    expfile = datadir("xstatt_" + f"{test_number:02d}" + "_OUT.pkl")
    dummy_test(infile, expfile)



"""Unit tests of linear_model."""
import pickle

import numpy as np
import pytest

from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect, MixedEffect
from brainstat.tests.testutil import datadir


def dummy_test(infile, expfile):

    ifile = open(infile, "br")
    Din = pickle.load(ifile)
    ifile.close()

    Y = Din["Y"]
    M = Din["M"]

    # Convert M to a true BrainStat model
    fixed_effects = FixedEffect(M[:, Din["n_random"] :])
    if Din["n_random"] != 0:
        mixed_effects = MixedEffect(
            M[:, : Din["n_random"]],
            name_ran=["f" + str(x) for x in range(Din["n_random"])],
        )
        M = fixed_effects + mixed_effects
    else:
        M = fixed_effects

    # assign slm params
    slm = SLM(M, FixedEffect(1), surf=Din["surf"])

    # here we go --> run the linear model
    slm._linear_model(Y)

    ofile = open(expfile, "br")
    Dout = pickle.load(ofile)
    ofile.close()

    # compare...
    testout = []

    for k, v in Dout.items():
        if k == "surf":
            # Surface data is only stored for reconstruction in MATLAB.
            continue

        a = getattr(slm, k)

        comp = np.allclose(a, v, rtol=1e-05, equal_nan=True)
        testout.append(comp)
    assert all(testout)


expected_number_of_tests = 16
parametrize = pytest.mark.parametrize


@parametrize("test_number", range(1, expected_number_of_tests + 1))
def test_run_all(test_number):
    infile = datadir("xlinmod_" + f"{test_number:02d}" + "_IN.pkl")
    expfile = datadir("xlinmod_" + f"{test_number:02d}" + "_OUT.pkl")
    dummy_test(infile, expfile)



"""Utilities for the stats functions."""

import warnings
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
from scipy.interpolate import interp1d

from brainstat._typing import ArrayLike


def row_ismember(a: np.ndarray, b: np.ndarray) -> List[Optional[int]]:
    """Tests whether rows of a occur in b.

    Parameters
    ----------
    a : numpy.array
        a 2D array with the same number of columns as b.
    b : numpy.array
        a 2D array with the same number of columns as a.

    Returns
    -------
    list
        Indices of rows in a that occur in b.
    """

    bind = {}
    for i, elt in enumerate(b):
        if tuple(elt) not in bind:
            bind[tuple(elt)] = i
    return [bind.get(tuple(itm), None) for itm in a]


def interp1(
    x: ArrayLike, y: ArrayLike, ix: ArrayLike, kind: Union[str, int] = "linear"
):
    """Interpolation between datapoints.

    Parameters
    ----------
    x : ArrayLike
        x coordinates of training data.
    y : ArrayLike
        y coordinates of training data.
    ix : ArrayLike
        coordinates of the interpolated points.
    kind str, int, optional
        type of interpolation; see scipy.interpolate.interp1d for options.

    Returns
    -------
    numpy.array
        interpolated y coordinates.
    """

    f = interp1d(x, y, kind, bounds_error=False, fill_value=np.nan)
    iy = f(ix)
    return iy


def ismember(
    A: np.ndarray, B: np.ndarray, rows: bool = False
) -> Tuple[bool, np.ndarray]:
    """Tests whether elements of A appear in B.

    Parameters
    ----------
    A : numpy.ndarray
        1D or 2D array
    B : numpy.ndarray
        1D or 2D array
    rows : bool, optional
        If true test for row correspondence rather than element correpondence.

    Returns
    -------
    bool
        Boolean of the same size as A denoting which elements (or rows) occur in B.
    numpy.ndarray
        Indices of matching elements/rows in A.

    Notes
    -----
    For row-wise comparisons, row_ismember should be significantly faster.

    """

    if rows:
        # Get rows of A that are in B.
        equality = np.equal(np.expand_dims(A, axis=2), np.expand_dims(B.T, axis=0))
        equal_rows = np.squeeze(np.all(equality, axis=1))
        bool_array = np.any(equal_rows, 1)

        # Get location of elements in B.
        locations = np.zeros(bool_array.shape) + np.nan
        for i in range(0, equal_rows.shape[0]):
            nz = np.nonzero(equal_rows[i, :])
            if nz[0].size != 0:
                locations[i] = nz[0]

    else:
        # Get values of A that are in B.
        bool_vector = np.in1d(A, B)
        bool_array = np.reshape(bool_vector, A.shape)

        # Get location of elements in B. Transpose B and A to get MATLAB behavior (i.e. column first)
        val, locB = np.unique(B.T, return_index=True)
        idx = np.flatnonzero(bool_array)
        locations = np.zeros(A.size) + np.nan
        Aflat = A.T.flat
        for i in range(0, idx.size):
            locations[idx[i]] = locB[np.argwhere(val == Aflat[idx[i]])]
        locations = np.reshape(locations, A.shape)
    return bool_array, locations


def colon(start: float, stop: float, increment: float = 1) -> np.ndarray:
    """Generates a range of numbers including the stop number.

    Parameters
    ----------
    start : float
        Starting number of the range.
    stop : float
        Stopping number of the range.
    increment : float, optional
        Increments of the range, defaults to 1.

    Returns
    -------
    numpy.array
        The requested numbers.
    """
    r = np.arange(start, stop, increment)
    if (start > stop and increment > 0) or (start < stop and increment < 0):
        return r
    elif start == stop or r[-1] + increment == stop:
        r = np.append(r, stop)
    return r


def apply_mask(Y: np.ndarray, mask: np.ndarray, axis: int = 0) -> np.ndarray:
    """Masks the data along a specified axis

    Parameters
    ----------
    Y : numpy.ndarray
        Data to be masked.
    mask : numpy.ndarray
        Boolean vector containing True for each element to keep.
    axis : int, optional
        Axis along which to operate, by default 0.

    Returns
    -------
    numpy.ndarray
        Masked data.
    """
    Y = Y.swapaxes(0, axis)
    Y = Y[mask, ...]
    return Y.swapaxes(0, axis)


def undo_mask(
    Y: np.ndarray, mask: np.ndarray, axis: int = 0, missing_value: float = np.nan
) -> np.ndarray:
    """Restores the original dimensions of masked data.

    Parameters
    ----------
    Y : numpy.ndarray
        Masked data.
    mask : numpy.ndarray
        Boolean vector used to mask the data.
    axis : int, optional
        Axis along which to operate, by default 0.
    missing_value : scalar, optional
        Number to insert for missing values, by default np.nan.

    Returns
    -------
    numpy.ndarray
        Unmasked data.
    """
    new_dims = list(Y.shape)
    new_dims[axis] = mask.size
    Y2 = np.empty(new_dims)
    Y2[:] = missing_value

    Y = Y.swapaxes(0, axis)
    Y2 = Y2.swapaxes(0, axis)
    Y2[mask, ...] = Y
    Y2 = Y2.swapaxes(0, axis)

    return Y2
